from braceexpand import braceexpand

from .base import (
    CheckpointableDataset,
    CheckpointableIterator,
    FileFormat,
    ParallelExecutorType,
    Sample,
    StateDict,
    TokenArray,
)
from .combinations import concat_datasets, interleave_datasets
from .version import __version__


__all__ = [
    "braceexpand",
    "CheckpointableDataset",
    "CheckpointableIterator",
    "FileFormat",
    "Sample",
    "StateDict",
    "TokenArray",
    "interleave_datasets",
    "ParallelExecutorType",
    "concat_datasets",
    "__version__",
]



from __future__ import annotations

import copy
from typing import Any, Optional, Sequence

import torch

from ...base import CheckpointableDataset, CheckpointableIterator, Sample, StateDict
from .tokenizer_utils import TokensQueue


class PackChunkIterator(CheckpointableIterator):
    def __init__(
        self,
        dataset: PackChunkDataset,
        source: CheckpointableIterator,
        buffers: Optional[dict[str, torch.Tensor]],
    ) -> None:
        self.dataset = dataset
        self.queue = TokensQueue(columns=self.dataset.target_columns, buffers=buffers)
        self.source = source

    def __next__(self) -> Sample:
        # Add samples until the buffer exceeds the chunk length
        while True:
            try:
                input_sample = next(self.source)
            except StopIteration:
                input_tensor_dict = None
                break
            input_tensor_dict = self.queue.tensor_dict_from_sample(input_sample)
            input_length = len(input_tensor_dict[self.dataset.target_columns[0]])

            # Discard too long samples
            if self.dataset.discard_long_samples and input_length > self.dataset.chunk_length:
                continue

            buffer_length = self.queue.length()

            if input_length + buffer_length > self.dataset.chunk_length:
                break
            else:
                self.queue.push_from_tensor_dict(input_tensor_dict)

        # Take the sample from the buffer, then add the remaining sample
        output_sample = self.queue.pop_all()
        if input_tensor_dict is None:
            output_length = len(output_sample[self.dataset.target_columns[0]])
            # If the source is exhausted and the buffer is empty, then we are done
            if output_length == 0:
                raise StopIteration()
        else:
            self.queue.push_from_tensor_dict(input_tensor_dict)

        # Truncate the output sample
        for column in self.dataset.target_columns:
            if len(output_sample[column]) > self.dataset.chunk_length:
                # Truncate
                assert not self.dataset.discard_long_samples  # This should have been discarded
                output_sample[column] = output_sample[column][: self.dataset.chunk_length]

        return output_sample

    def state_dict(self) -> StateDict:
        return {
            "source": self.source.state_dict(),
            "buffers": self.queue.buffers.copy(),
        }


class PackChunkDataset(CheckpointableDataset):
    def __init__(
        self,
        source: CheckpointableDataset,
        chunk_length: int,
        target_columns: Sequence[str],
        discard_long_samples: bool = False,
    ) -> None:
        self.source = source
        self.target_columns = target_columns
        self.chunk_length = chunk_length
        self.discard_long_samples = discard_long_samples

    def iter(self, state_dict: Optional[dict[str, Any]] = None) -> CheckpointableIterator:
        if state_dict:
            source_state_dict = state_dict.pop("source")
            buffers = copy.copy(state_dict.pop("buffers"))
            if state_dict:
                raise ValueError(f"Unexpected keys in state_dict: {state_dict.keys()}")
        else:
            source_state_dict = None
            buffers = None

        source = self.source.iter(state_dict=source_state_dict)
        return PackChunkIterator(self, source, buffers)



from __future__ import annotations

import copy
from typing import Any, Optional, Sequence

import torch

from ...base import CheckpointableDataset, CheckpointableIterator, Sample, StateDict
from .tokenizer_utils import TokensQueue


class ConcatChunkIterator(CheckpointableIterator):
    def __init__(
        self,
        dataset: ConcatChunkDataset,
        source: CheckpointableIterator,
        buffers: Optional[dict[str, torch.Tensor]],
    ) -> None:
        self.dataset = dataset
        self.source = source
        self.queue = TokensQueue(columns=self.dataset.target_columns, buffers=buffers)

    def __next__(self) -> Sample:
        while self.queue.length() < self.dataset.chunk_length:
            in_sample = next(self.source)
            self.queue.push_from_sample(in_sample)

        return self.queue.pop_by_length(self.dataset.chunk_length)

    def state_dict(self) -> StateDict:
        return {
            "source": self.source.state_dict(),
            "buffers": self.queue.buffers.copy(),
        }


class ConcatChunkDataset(CheckpointableDataset):
    def __init__(
        self,
        source: CheckpointableDataset,
        chunk_length: int,
        target_columns: Sequence[str],
    ) -> None:
        self.source = source
        self.target_columns = target_columns
        self.chunk_length = chunk_length

    def iter(self, state_dict: Optional[dict[str, Any]] = None) -> CheckpointableIterator:
        if state_dict:
            source_state_dict = state_dict.pop("source")
            buffers = copy.copy(state_dict.pop("buffers"))
            if state_dict:
                raise ValueError(f"Unexpected keys in state_dict: {state_dict.keys()}")
        else:
            source_state_dict = None
            buffers = None

        source = self.source.iter(state_dict=source_state_dict)
        return ConcatChunkIterator(self, source, buffers)



from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING, Optional

import torch

from ...base import CheckpointableDataset, Sample
from .tokenizer_utils import tensor_from_token_array


if TYPE_CHECKING:
    from transformers import PreTrainedTokenizerBase


TOKENIZATION_EXAMPLE_STR = "Hello world!"


@dataclasses.dataclass
class TokenizerBehavior:
    bos_token: Optional[str]
    eos_token: Optional[str]
    bos_token_id: Optional[int]
    eos_token_id: Optional[int]

    # Some tokenizers (like "gpt-neox-20b" and "gpt2") have the same token for bos and eos.
    bos_eos_equal: bool

    # Tokenization result for `TOKENIZATION_EXAMPLE_STR`
    ids_example: list[int]

    # Does the tokenizer add the BOS and EOS tokens automatically?
    bos_token_added: bool
    eos_token_added: bool

    def from_tokenizer(tokenizer: PreTrainedTokenizerBase) -> TokenizerBehavior:
        ids_example = tokenizer.encode(TOKENIZATION_EXAMPLE_STR)
        if not isinstance(ids_example, list):
            raise ValueError(f"Tokenizer returned {type(ids_example)}, expected list")
        if len(ids_example) == 0:
            raise ValueError(
                f"Tokenizer returned empty list for input string: {TOKENIZATION_EXAMPLE_STR}"
            )

        if tokenizer(TOKENIZATION_EXAMPLE_STR)["input_ids"] != ids_example:
            raise ValueError("Tokenizer returned different results for `encode` and `__call__`")

        if len(ids_example) > 0 and ids_example[0] == tokenizer.bos_token_id:
            bos_token_added = True
        else:
            bos_token_added = False

        if len(ids_example) > 0 and ids_example[-1] == tokenizer.eos_token_id:
            eos_token_added = True
        else:
            eos_token_added = False

        return TokenizerBehavior(
            bos_token=tokenizer.bos_token,
            eos_token=tokenizer.eos_token,
            bos_token_id=tokenizer.bos_token_id,
            eos_token_id=tokenizer.eos_token_id,
            bos_eos_equal=(tokenizer.bos_token_id == tokenizer.eos_token_id),
            ids_example=ids_example,
            bos_token_added=bos_token_added,
            eos_token_added=eos_token_added,
        )


def add_bos_eos(
    source: CheckpointableDataset,
    bos_token_id: Optional[int],
    eos_token_id: Optional[int],
    target_column: str,
) -> CheckpointableDataset:
    if bos_token_id is not None:
        bos_tensor = tensor_from_token_array([bos_token_id])
    else:
        bos_tensor = tensor_from_token_array([])

    if eos_token_id is not None:
        eos_tensor = tensor_from_token_array([eos_token_id])
    else:
        eos_tensor = tensor_from_token_array([])

    def _fn(sample: Sample) -> Sample:
        sample = sample.copy()
        tokens = tensor_from_token_array(sample[target_column])
        sample[target_column] = torch.cat((bos_tensor, tokens, eos_tensor))
        return sample

    return source.map(_fn)


def ensure_bos_eos(
    source: CheckpointableDataset,
    tokenizer: PreTrainedTokenizerBase,
    target_column: str,
) -> CheckpointableDataset:
    behavior = TokenizerBehavior.from_tokenizer(tokenizer)

    # Some tokenizers (like "gpt2" and "gpt-neox-20b") have the same token for bos and eos.
    # In this case, we don't have to add both.
    if behavior.bos_eos_equal:
        if behavior.bos_token_added or behavior.eos_token_added:
            # Either of BOS or EOS is already added, so we don't need to add anything.
            return source
        else:
            # Either of BOS or EOS should be added, but not both.
            # Here, we choose to add EOS.
            return add_bos_eos(source, None, tokenizer.eos_token_id, target_column)
    else:
        if not behavior.bos_token_added and tokenizer.bos_token:
            bos_token_to_add = tokenizer.bos_token_id
        else:
            bos_token_to_add = None

        if not behavior.eos_token_added and tokenizer.eos_token:
            eos_token_to_add = tokenizer.eos_token_id
        else:
            bos_token_to_add = None

        return add_bos_eos(source, bos_token_to_add, eos_token_to_add, target_column)



from __future__ import annotations

from typing import Optional

from ...base import CheckpointableDataset, CheckpointableIterator, Sample, StateDict


class StrideIterator(CheckpointableIterator):
    def __init__(
        self, source: CheckpointableIterator, interval: int, offset: int, start_count: int
    ) -> None:
        self.source = source
        self.interval = interval
        self.offset = offset
        self.count = start_count

    def __next__(self) -> Sample:
        while self.count % self.interval != self.offset:
            self.count += 1
            next(self.source)

        self.count += 1
        return next(self.source)

    def state_dict(self) -> StateDict:
        return {
            "source": self.source.state_dict(),
            "count": self.count,
        }


class StrideDataset(CheckpointableDataset):
    def __init__(self, source: CheckpointableDataset, interval: int, offset: int) -> None:
        if interval <= 0:
            raise ValueError(f"interval must be positive: {interval}")
        if offset < 0 or offset >= interval:
            raise ValueError(f"offset must be in [0, interval): {offset}")

        self.source = source
        self.interval = interval
        self.offset = offset

    def iter(self, state_dict: Optional[StateDict] = None) -> CheckpointableIterator:
        if state_dict is not None:
            source = state_dict.pop("source")
            count = state_dict.pop("count")
        else:
            source = None
            count = 0

        return StrideIterator(
            source=self.source.iter(source),
            interval=self.interval,
            offset=self.offset,
            start_count=count,
        )



from __future__ import annotations

import itertools
import multiprocessing
import os
import threading
from logging import getLogger
from multiprocessing import Process, Queue
from multiprocessing.context import ForkServerProcess, SpawnProcess
from threading import Thread
from typing import Any, Generator, Iterator, Optional, Type, Union

from ...base import (
    CheckpointableDataset,
    CheckpointableIterator,
    FilterMapFn,
    ParallelExecutorType,
    Sample,
    StateDict,
)


logger = getLogger(__name__)
mp_ctx = multiprocessing.get_context("forkserver")  # TODO: customization?


class StopToken:
    pass


WorkerClass = Union[Type[Process], Type[Thread], Type[ForkServerProcess], Type[SpawnProcess]]
WorkerInput = Union[Sample, StopToken]


class WorkerResult:
    error: Optional[Exception]
    result: Optional[Sample]

    def __init__(self, error: Optional[Exception], result: Optional[Sample]) -> None:
        self.error = error
        self.result = result


def _get_worker_class(executor_type: str) -> WorkerClass:
    if executor_type == "process":
        return mp_ctx.Process
    elif executor_type == "thread":
        return Thread
    else:
        raise ValueError('Invalid executor_type. Choose either "process" or "thread".')


def _worker(fn: FilterMapFn, rx: Queue[WorkerInput], tx: Queue[WorkerResult]) -> None:
    process_id = os.getpid()
    thread_id = threading.get_ident()
    logger.debug(f"Worker starting. Process ID: {process_id}, Thread ID: {thread_id}.")

    while True:
        try:
            item = rx.get()
            if isinstance(item, StopToken):
                break

            result = fn(item)
            tx.put(WorkerResult(error=None, result=result))
        except Exception as e:
            tx.put(WorkerResult(error=e, result=None))

    logger.debug(f"Worker ending. Process ID: {process_id}, Thread ID: {thread_id}.")


def _imap_ordered(
    fn: FilterMapFn,
    source: Iterator[Sample],
    max_workers: int,
    queue_len: int,
    executor_type: str,
) -> Generator[Optional[Sample], None, None]:
    worker_class = _get_worker_class(executor_type)

    workers = []
    try:
        logger.debug(f"Starting {max_workers} workers. This may take some time.")
        for _ in range(max_workers):
            worker_queue_len = queue_len // max_workers + 1
            rx: Queue[WorkerInput] = mp_ctx.Queue(worker_queue_len)
            tx: Queue[WorkerResult] = mp_ctx.Queue(worker_queue_len)
            worker = worker_class(target=_worker, args=(fn, rx, tx), daemon=True)
            worker.start()
            workers.append((worker, rx, tx))
        logger.debug(f"Started {max_workers} workers.")

        # Sample index to read and put
        get_index = 0
        put_index = 0

        # Fill the queues
        for x in itertools.islice(source, queue_len):
            _, rx, _ = workers[put_index % max_workers]
            rx.put(x)
            put_index += 1

        # Read from the queues
        while get_index < put_index:
            _, _, tx = workers[get_index % max_workers]
            result = tx.get()
            if result.error:
                raise result.error
            yield result.result
            get_index += 1

            _, rx, _ = workers[put_index % max_workers]
            try:
                rx.put(next(source))
                put_index += 1
            except StopIteration:
                # All tasks submitted, just wait for them to complete
                continue
    finally:
        logger.debug(f"Terminating {max_workers} workers.")
        if executor_type == "process":
            for worker, _, _ in workers:
                assert not isinstance(worker, Thread)
                worker.terminate()
        else:
            # Thread does not have terminate method
            for _, rx, _ in workers:
                rx.put(StopToken())
            for worker, _, _ in workers:
                worker.join()

        logger.debug(f"Terminated {max_workers} workers.")


def _imap_unordered(
    fn: FilterMapFn,
    source: Iterator[Sample],
    max_workers: int,
    queue_len: int,
    executor_type: str,
) -> Generator[Optional[Sample], None, None]:
    worker_class = _get_worker_class(executor_type)
    rx: Queue[WorkerInput] = mp_ctx.Queue(queue_len + 1)
    tx: Queue[WorkerResult] = mp_ctx.Queue(queue_len + 1)

    workers = []
    try:
        logger.debug(f"Starting {max_workers} workers. This may take some time.")
        for _ in range(max_workers):
            worker = worker_class(target=_worker, args=(fn, rx, tx))
            worker.start()
            workers.append((worker, rx, tx))
        logger.debug(f"Started {max_workers} workers.")

        get_index = 0
        put_index = 0

        # Fill the queues
        for x in itertools.islice(source, queue_len):
            rx.put(x)
            put_index += 1

        # Read from the queues
        while get_index < put_index:
            result = tx.get()
            if result.error:
                raise result.error
            yield result.result
            get_index += 1

            try:
                rx.put(next(source))
                put_index += 1
            except StopIteration:
                # All tasks submitted, just wait for them to complete
                continue
    finally:
        logger.debug(f"Terminating {max_workers} workers.")
        if executor_type == "process":
            for worker, _, _ in workers:
                assert not isinstance(worker, Thread)
                worker.terminate()
        else:
            # Thread does not have terminate method
            for _, rx, _ in workers:
                rx.put(StopToken())
            for worker, _, _ in workers:
                worker.join()

        logger.debug(f"Terminated {max_workers} workers.")


class ParallelFilterMapIterator(CheckpointableIterator):
    def __init__(
        self,
        source: CheckpointableIterator,
        dataset: ParallelFilterMapDataset,
        unconsumed_outputs: list[Sample],
    ) -> None:
        self.source = source
        self.dataset = dataset
        self._closing = False
        self._iter = self._start(unconsumed_outputs)

    def _source_iter(self) -> Generator[Sample, None, None]:
        while not self._closing:
            try:
                yield next(self.source)
            except StopIteration:
                break

    def _output_iter(self, unconsumed_outputs: list[Sample]) -> Generator[Sample, None, None]:
        # `self.closing = False` should not be in this method; it should be done before calling
        # this method. This is because the execution is delayed until the first `next()` call.
        # In the edge case where `state_dict` (and thus `_close`) is called before `next`,
        # `_source_iter` cannot be closed properly.

        if self.dataset.ordered:
            it = _imap_ordered(
                self.dataset.fn,
                self._source_iter(),
                self.dataset.max_workers,
                self.dataset.queue_len,
                self.dataset.executor_type,
            )
        else:
            it = _imap_unordered(
                self.dataset.fn,
                self._source_iter(),
                self.dataset.max_workers,
                self.dataset.queue_len,
                self.dataset.executor_type,
            )
        yield from itertools.chain(unconsumed_outputs, it)

    def __next__(self) -> Sample:
        while True:
            sample = next(self._iter)
            if sample is not None:
                return sample

    def _start(self, unconsumed_outputs: list[Sample]) -> Generator[Sample, None, None]:
        self._closing = False
        return self._output_iter(unconsumed_outputs)

    def _close(self) -> list[Sample]:
        self._closing = True
        return list(self._iter)

    def state_dict(self) -> StateDict:
        unconsumed_outputs = self._close()
        state_dict = {
            "source": self.source.state_dict(),
            "unconsumed_outputs": unconsumed_outputs,
        }
        self._closing = False
        self._iter = self._output_iter(unconsumed_outputs)
        return state_dict


def _get_default_max_workers() -> int:
    n_cpus = os.cpu_count() or 1
    local_world_size = int(os.environ.get("LOCAL_WORLD_SIZE", 1))
    return max(n_cpus // local_world_size, 1)


class ParallelFilterMapDataset(CheckpointableDataset):
    def __init__(
        self,
        source: CheckpointableDataset,
        fn: FilterMapFn,
        max_workers: Optional[int],
        prefetch_factor: int,
        ordered: bool,
        executor_type: ParallelExecutorType,
    ):
        self.source = source
        self.fn = fn

        if max_workers is None:
            self.max_workers = _get_default_max_workers()
        else:
            self.max_workers = max_workers
        if self.max_workers < 1:
            raise ValueError("max_workers must be greater than 0: {self.max_workers}")

        if prefetch_factor < 1:
            raise ValueError("prefetch_factor must be greater than 0: {prefetch_factor}")
        self.queue_len = self.max_workers * prefetch_factor

        self.ordered = ordered
        self.executor_type = executor_type

    def iter(self, state_dict: Optional[dict[str, Any]] = None) -> CheckpointableIterator:
        if state_dict is not None:
            unconsumed_outputs = state_dict.pop("unconsumed_outputs")
            source_state_dict = state_dict.pop("source")
            if state_dict:
                raise ValueError(f"Unexpected keys in state_dict: {state_dict.keys()}")
        else:
            unconsumed_outputs = []
            source_state_dict = None
        iter = self.source.iter(state_dict=source_state_dict)
        return ParallelFilterMapIterator(iter, self, unconsumed_outputs)



from __future__ import annotations

import time
from logging import getLogger

from ...base import CheckpointableIterator, FileFormat, Sample, StateDict
from .generators import yield_samples


logger = getLogger()


class ShardReader(CheckpointableIterator):
    def __init__(
        self,
        url: str,
        format: FileFormat,
        timeout: float,
        n_prefetch_samples: int,
        n_samples_yielded: int,
        epoch: int,
        index_in_epoch: int,
        max_retries: int = 10,
        wait_time_base: float = 1.0,
        wait_time_max: float = 600.0,
        wait_time_multiplier: float = 2.0,
    ):
        self.url = url
        self.format = format
        self.timeout = timeout
        self.n_prefetch_samples = n_prefetch_samples
        self.epoch = epoch
        self.index_in_epoch = index_in_epoch
        self.max_retries = max_retries
        self.wait_time_base = wait_time_base
        self.wait_time_max = wait_time_max
        self.wait_time_multiplier = wait_time_multiplier

        self.n_samples_yielded = n_samples_yielded
        self.iter = yield_samples(
            self.url,
            self.format,
            self.n_samples_yielded,
            self.n_prefetch_samples,
            self.timeout,
        )

    def __next__(self) -> Sample:
        # For exponential backoff
        wait_time = self.wait_time_base

        n_retries = 0
        while n_retries < self.max_retries:
            try:
                sample = next(self.iter)
                self.n_samples_yielded += 1
                wait_time = self.wait_time_base  # Reset exponential backoff
                return sample
            except StopIteration:
                raise
            except Exception:
                n_retries += 1
                logger.exception(
                    f"Error while reading from {self.url}. "
                    f"Retry {n_retries}/{self.max_retries} in {wait_time} seconds."
                )
                time.sleep(wait_time)
                wait_time = min(self.wait_time_max, wait_time * self.wait_time_multiplier)

                self.iter = yield_samples(
                    self.url,
                    self.format,
                    self.n_samples_yielded,
                    self.n_prefetch_samples,
                    self.timeout,
                )

        raise Exception(f"Failed to read from {self.url} after {self.max_retries} attempts.")

    def state_dict(self) -> StateDict:
        return {
            "url": self.url,
            "format": self.format,
            "n_samples_yielded": self.n_samples_yielded,
            "epoch": self.epoch,
            "index_in_epoch": self.index_in_epoch,
        }

    @staticmethod
    def from_state_dict(
        state_dict: StateDict,
        timeout: float,
        n_prefetch_samples: int,
    ) -> ShardReader:
        url = state_dict.pop("url")
        format = state_dict.pop("format")
        n_samples_yielded = state_dict.pop("n_samples_yielded")
        epoch = state_dict.pop("epoch")
        index_in_epoch = state_dict.pop("index_in_epoch")
        if state_dict:
            raise ValueError(f"Unused keys in state_dict: {state_dict.keys()}")

        return ShardReader(
            url=url,
            format=format,
            n_samples_yielded=n_samples_yielded,
            timeout=timeout,
            n_prefetch_samples=n_prefetch_samples,
            epoch=epoch,
            index_in_epoch=index_in_epoch,
        )



from .bos_eos import add_bos_eos, ensure_bos_eos
from .chunk import ChunkDataset, ChunkIterator
from .concat_chunk import ConcatChunkDataset, ConcatChunkIterator
from .pack_chunk import PackChunkDataset, PackChunkIterator
from .padding import pad
from .tokenization import tokenize
from .tokenizer_utils import tensor_from_token_array


__all__ = [
    "tensor_from_token_array",
    "ChunkDataset",
    "ChunkIterator",
    "ConcatChunkDataset",
    "ConcatChunkIterator",
    "PackChunkDataset",
    "PackChunkIterator",
    "pad",
    "tokenize",
    "add_bos_eos",
    "ensure_bos_eos",
]



from __future__ import annotations

import functools
from typing import Any, Optional

from ...base import (
    CheckpointableDataset,
    CheckpointableIterator,
    FilterFn,
    FilterMapFn,
    MapFn,
    Sample,
    StateDict,
)


class FilterMapIterator(CheckpointableIterator):
    def __init__(self, source: CheckpointableIterator, fn: FilterMapFn):
        self.source = source
        self.fn = fn

    def __next__(self) -> Sample:
        while True:
            sample = self.fn(next(self.source))
            if sample is not None:
                return sample

    def state_dict(self) -> StateDict:
        return self.source.state_dict()


class FilterMapDataset(CheckpointableDataset):
    def __init__(self, source: CheckpointableDataset, fn: FilterMapFn):
        self.source = source
        self.fn = fn

    def iter(self, state_dict: Optional[dict[str, Any]] = None) -> CheckpointableIterator:
        iter = self.source.iter(state_dict=state_dict)
        return FilterMapIterator(iter, self.fn)


def _map_fn_adapter(sample: Sample, map_fn: MapFn) -> Optional[Sample]:
    sample = map_fn(sample)
    if sample is None:
        raise ValueError("map_fn must not return None")
    return sample


def _filter_fn_adapter(sample: Sample, filter_fn: FilterFn) -> Optional[Sample]:
    return sample if filter_fn(sample) else None


def adapt_map_fn(map_fn: MapFn) -> FilterMapFn:
    return functools.partial(_map_fn_adapter, map_fn=map_fn)


def adapt_filter_fn(filter_fn: FilterFn) -> FilterMapFn:
    return functools.partial(_filter_fn_adapter, filter_fn=filter_fn)



from __future__ import annotations

import itertools
import random
from typing import Callable

import numpy as np
import torch

from .base import CheckpointableDataset, Sample, TokenArray


def generate_example_sequence(n_samples: int = 20) -> list[Sample]:
    return [{"id": i} for i in range(n_samples)]


def sort_key_fn(sample: Sample) -> int:
    """Key function for sorting samples by their id"""
    return sample["id"]  # type: ignore


def generate_tokenized_samples(
    min_length: int = 0,
    max_length: int = 10,
    n_examples: int = 20,
    vocab_size: int = 100,
    seed: int = 42,
) -> list[dict[str, TokenArray]]:
    rng = random.Random(seed)
    i = 0
    samples = []
    for _ in range(n_examples):
        seq_len = rng.randint(min_length, max_length)
        seq = []
        for _ in range(seq_len):
            seq.append(i % vocab_size)
            i += 1
        sample: dict[str, TokenArray] = {"input_ids": seq}
        samples.append(sample)
    return samples


def assert_samples_equal(sample1: Sample, sample2: Sample) -> None:
    assert sample1.keys() == sample2.keys(), "Samples' keys are different"
    for key in sample1.keys():
        item1 = sample1[key]
        item2 = sample2[key]
        if isinstance(item1, torch.Tensor) and isinstance(item2, torch.Tensor):
            assert torch.equal(item1, item2), f"Tensors do not match for key {key}"
        elif isinstance(item1, np.ndarray) and isinstance(item2, np.ndarray):
            assert np.array_equal(item1, item2), f"Arrays do not match for key {key}"
        else:
            assert item1 == item2, f"Items do not match for key {key}"


def assert_sample_lists_equal(samples1: list[Sample], samples2: list[Sample]) -> None:
    assert len(samples1) == len(samples2), "Samples have different lengths"
    for sample1, sample2 in zip(samples1, samples2):
        assert_samples_equal(sample1, sample2)


def check_resumption_with_instantiation(
    dataset_factory: Callable[[], CheckpointableDataset],
    ckpt_index: int,
    n_subsequent_samples: int = 10,
) -> None:
    # Consuming the first `ckpt_index` samples and get the state dict
    dataset = dataset_factory()
    it = iter(dataset)
    for _ in range(ckpt_index):
        next(it)
    state_dict = it.state_dict()
    subsequent_samples_original = list(itertools.islice(it, n_subsequent_samples))

    # Clean up the dataset
    del dataset, it

    # Resume from the state dict
    dataset = dataset_factory()
    it = dataset.iter(state_dict=state_dict)
    subsequent_samples_resumed = list(itertools.islice(it, n_subsequent_samples))

    assert_sample_lists_equal(subsequent_samples_original, subsequent_samples_resumed)


def check_resumption(
    original_dataset: CheckpointableDataset,
    resuming_dataset: CheckpointableDataset,
    ckpt_index: int,
    n_subsequent_samples: int = 10,
) -> None:
    # Consuming the first `ckpt_index` samples and get the state dict
    it = iter(original_dataset)
    for _ in range(ckpt_index):
        next(it)
    state_dict = it.state_dict()
    subsequent_samples_original = list(itertools.islice(it, n_subsequent_samples))

    # Resume from the state dict
    it = resuming_dataset.iter(state_dict=state_dict)
    subsequent_samples_resumed = list(itertools.islice(it, n_subsequent_samples))

    # print(subsequent_samples_original)
    # print(subsequent_samples_resumed)
    assert_sample_lists_equal(subsequent_samples_original, subsequent_samples_resumed)



from __future__ import annotations

import functools
import os
from typing import TYPE_CHECKING, Any, Dict, Optional

from ...base import CheckpointableDataset, ParallelExecutorType, Sample


if TYPE_CHECKING:
    from ...base import Tokenizer


def _map_fn(
    sample: Sample, tokenizer: Tokenizer, target_column: str, tokenizer_kwargs: dict[str, Any]
) -> Sample:
    sample = sample.copy()
    sample.update(tokenizer(sample[target_column], **tokenizer_kwargs))
    return sample


def tokenize(
    source: CheckpointableDataset,
    tokenizer: Tokenizer,
    tokenizer_kwargs: Optional[Dict[str, Any]],
    target_column: str,
    parallel: bool,
    max_workers: Optional[int],
    prefetch_factor: int,
    ordered: bool,
    executor_type: ParallelExecutorType,
) -> CheckpointableDataset:
    fn = functools.partial(
        _map_fn,
        tokenizer=tokenizer,
        target_column=target_column,
        tokenizer_kwargs=tokenizer_kwargs or {},
    )

    if parallel:
        # TODO: show some warning on this
        os.environ["TOKENIZERS_PARALLELISM"] = "false"

        return source.parallel_map(
            fn,
            max_workers=max_workers,
            prefetch_factor=prefetch_factor,
            ordered=ordered,
            executor_type=executor_type,
        )
    else:
        return source.map(fn)



from __future__ import annotations

import abc
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Union,
)

import numpy as np
import torch


if TYPE_CHECKING:
    import streaming
    from transformers import PreTrainedTokenizer, PreTrainedTokenizerFast

    Tokenizer = Union[PreTrainedTokenizer, PreTrainedTokenizerFast]


Sample = Dict[str, Any]
StateDict = Dict[str, Any]
TokenArray = Union[List[int], np.ndarray, torch.Tensor]
FilterMapFn = Callable[[Sample], Optional[Sample]]
MapFn = Callable[[Sample], Sample]
FilterFn = Callable[[Sample], bool]
CollateFn = Callable[[List[Sample]], Sample]
ParallelExecutorType = Literal["process", "thread"]
FileFormat = Literal["auto", "jsonl", "cbor"]


class CheckpointableIterator(abc.ABC):
    def __iter__(self) -> CheckpointableIterator:
        return self

    @abc.abstractmethod
    def __next__(self) -> Sample:
        raise NotImplementedError()

    @abc.abstractmethod
    def state_dict(self) -> StateDict:
        raise NotImplementedError


class CheckpointableDataset(torch.utils.data.IterableDataset, abc.ABC):
    def __iter__(self) -> CheckpointableIterator:
        return self.iter(state_dict=None)

    @abc.abstractmethod
    def iter(self, state_dict: Optional[StateDict] = None) -> CheckpointableIterator:
        raise NotImplementedError()

    @staticmethod
    def from_sequence(
        sequence: Sequence[Sample],
        repeat: bool = False,
        shuffle: bool = False,
        shuffle_seed: int = 42,
    ) -> CheckpointableDataset:
        from .sources import SequenceDataset

        return SequenceDataset(
            sequence=sequence, repeat=repeat, shuffle=shuffle, shuffle_seed=shuffle_seed
        )

    @staticmethod
    def from_iterable(
        iterable: Iterable[Sample],
        repeat: bool = False,
    ) -> CheckpointableDataset:
        """
        Create a CheckpointableDataset from an iterable.

        This static method creates a new CheckpointableDataset instance from an iterable.
        Each item in the iterable should be a `Sample` instance.

        The iterable should be a 're-iterable' and 'deterministic', i.e., it should return a new
        iterator each time `iter` is invoked and generate the same sequence of samples every time.
        This allows the successful resumption.
        """

        from .sources import IterableDataset

        return IterableDataset(iterable, repeat=repeat)

    @staticmethod
    def from_mosaicml(
        mosaicml_dataset: streaming.StreamingDataset,
        repeat: bool = False,
    ) -> CheckpointableDataset:
        from .sources import MosaicmlDataset

        return MosaicmlDataset(mosaicml_dataset, repeat=repeat)

    @staticmethod
    def from_files(
        urls: Union[str, Sequence[str]],
        repeat: bool = False,
        shuffle_shards: bool = False,
        format: FileFormat = "auto",
        n_active_shards: int = 10,
        n_standby_shards: int = 2,
        timeout: float = 60.0,
        n_prefetch_samples: int = 10,
        seed: int = 42,
    ) -> CheckpointableDataset:
        from .sources import FilesDataset

        return FilesDataset(
            urls=urls,
            repeat=repeat,
            shuffle_shards=shuffle_shards,
            format=format,
            n_active_shards=n_active_shards,
            n_standby_shards=n_standby_shards,
            timeout=timeout,
            n_prefetch_samples=n_prefetch_samples,
            seed=seed,
        )

    def filter_map(self, fn: FilterMapFn) -> CheckpointableDataset:
        from .transforms import FilterMapDataset

        return FilterMapDataset(self, fn)

    def map(self, fn: Callable[[Sample], Sample]) -> CheckpointableDataset:
        from .transforms import FilterMapDataset

        return FilterMapDataset(self, fn)

    def filter(self, fn: Callable[[Sample], bool]) -> CheckpointableDataset:
        from .transforms import FilterMapDataset

        def _fn(sample: Sample) -> Optional[Sample]:
            return sample if fn(sample) else None

        return FilterMapDataset(self, _fn)

    def parallel_filter_map(
        self,
        fn: FilterMapFn,
        max_workers: Optional[int] = None,
        prefetch_factor: int = 10,
        ordered: bool = True,
        executor_type: ParallelExecutorType = "process",
    ) -> CheckpointableDataset:
        from .transforms import ParallelFilterMapDataset

        return ParallelFilterMapDataset(
            self,
            fn,
            max_workers=max_workers,
            prefetch_factor=prefetch_factor,
            ordered=ordered,
            executor_type=executor_type,
        )

    def parallel_map(
        self,
        fn: Callable[[Sample], Sample],
        max_workers: Optional[int] = None,
        prefetch_factor: int = 10,
        ordered: bool = True,
        executor_type: ParallelExecutorType = "process",
    ) -> CheckpointableDataset:
        from .transforms.basic.filter_map import adapt_map_fn

        return self.parallel_filter_map(
            adapt_map_fn(fn),
            max_workers=max_workers,
            prefetch_factor=prefetch_factor,
            ordered=ordered,
            executor_type=executor_type,
        )

    def parallel_filter(
        self,
        fn: FilterFn,
        max_workers: Optional[int] = None,
        prefetch_factor: int = 10,
        ordered: bool = True,
        executor_type: ParallelExecutorType = "process",
    ) -> CheckpointableDataset:
        from .transforms.basic.filter_map import adapt_filter_fn

        return self.parallel_filter_map(
            adapt_filter_fn(fn),
            max_workers=max_workers,
            prefetch_factor=prefetch_factor,
            ordered=ordered,
            executor_type=executor_type,
        )

    def enumerate(self, count_column: str = "step") -> CheckpointableDataset:
        from .transforms import CountDataset

        return CountDataset(self, count_column=count_column)

    def take(
        self,
        max_count: int,
    ) -> CheckpointableDataset:
        from .transforms import CountDataset

        return CountDataset(self, max_count=max_count)

    def shuffle(
        self,
        buffer_size: int,
        seed: int = 42,
    ) -> CheckpointableDataset:
        from .transforms import ShuffleDataset

        return ShuffleDataset(self, buffer_size=buffer_size, seed=seed)

    def batch(
        self,
        batch_size: int,
        collate_fn: CollateFn = torch.utils.data.default_collate,
        drop_last: bool = False,
    ) -> CheckpointableDataset:
        from .transforms import BatchDataset

        return BatchDataset(
            self, batch_size=batch_size, collate_fn=collate_fn, drop_last=drop_last
        )

    def stride(
        self,
        interval: int,
        offset: int,
    ) -> CheckpointableDataset:
        from .transforms import StrideDataset

        return StrideDataset(self, interval=interval, offset=offset)

    def tokenize(
        self,
        tokenizer: Tokenizer,
        tokenizer_kwargs: Optional[Dict[str, Any]] = None,
        target_column: str = "text",
        parallel: bool = True,
        max_workers: Optional[int] = 1,
        prefetch_factor: int = 10,
        ordered: bool = True,
        executor_type: ParallelExecutorType = "thread",
    ) -> CheckpointableDataset:
        from .transforms import tokenize

        return tokenize(
            source=self,
            tokenizer=tokenizer,
            tokenizer_kwargs=tokenizer_kwargs,
            target_column=target_column,
            parallel=parallel,
            max_workers=max_workers,
            prefetch_factor=prefetch_factor,
            ordered=ordered,
            executor_type=executor_type,
        )

    def add_bos_eos(
        self,
        bos_token_id: Optional[int],
        eos_token_id: Optional[int],
        target_column: str = "input_ids",
    ) -> CheckpointableDataset:
        from .transforms import add_bos_eos

        return add_bos_eos(
            self, bos_token_id=bos_token_id, eos_token_id=eos_token_id, target_column=target_column
        )

    def ensure_bos_eos(
        self, tokenizer: Tokenizer, target_column: str = "input_ids"
    ) -> CheckpointableDataset:
        from .transforms import ensure_bos_eos

        return ensure_bos_eos(self, tokenizer=tokenizer, target_column=target_column)

    def pad(
        self,
        pad_values: Mapping[str, int],
        chunk_length: int,
    ) -> CheckpointableDataset:
        from .transforms import pad

        return pad(self, pad_values=pad_values, chunk_length=chunk_length)

    def chunk(
        self,
        chunk_length: int,
        target_columns: Sequence[str] = ("input_ids",),
        drop_remainder: bool = True,
    ) -> CheckpointableDataset:
        from .transforms import ChunkDataset

        return ChunkDataset(
            self,
            chunk_length=chunk_length,
            target_columns=target_columns,
            drop_remainder=drop_remainder,
        )

    def concat_chunk(
        self,
        chunk_length: int,
        target_columns: Sequence[str] = ("input_ids",),
    ) -> CheckpointableDataset:
        from .transforms import ConcatChunkDataset

        return ConcatChunkDataset(
            self,
            chunk_length=chunk_length,
            target_columns=target_columns,
        )

    def pack_chunk(
        self,
        chunk_length: int,
        target_columns: Sequence[str],
        discard_long_samples: bool = False,
    ) -> CheckpointableDataset:
        from .transforms import PackChunkDataset

        return PackChunkDataset(
            self,
            chunk_length=chunk_length,
            target_columns=target_columns,
            discard_long_samples=discard_long_samples,
        )

    # `__add__` is implemented in PyTorch's `IterableDataset`,
    # so we need to override it here for prevent unexpected behavior
    def __add__(self, other: CheckpointableDataset) -> CheckpointableDataset:  # type: ignore
        from .combinations import concat_datasets

        return concat_datasets([self, other])



from .files import FilesDataset
from .iterable import IterableDataset
from .mosaicml import MosaicmlDataset
from .sequence import SequenceDataset


__all__ = [
    "FilesDataset",
    "SequenceDataset",
    "IterableDataset",
    "MosaicmlDataset",
]



from __future__ import annotations

import functools
from typing import Mapping

import torch

from ...base import CheckpointableDataset, Sample
from .tokenizer_utils import tensor_from_token_array


def _pad_fn(
    sample: Sample,
    pad_values: Mapping[str, int],
    chunk_length: int,
) -> Sample:
    sample = sample.copy()
    for column, pad_value in pad_values.items():
        if column not in sample:
            raise ValueError(f"Column {column} not found in sample (columns: {sample.keys()})")

        input_tensor = tensor_from_token_array(sample[column])
        input_length = len(input_tensor)
        if input_length > chunk_length:
            raise ValueError(f"Input sequence too long ({input_length} > {chunk_length})")

        pad_length = chunk_length - input_length
        pad_tensor = torch.full((pad_length,), pad_value, dtype=torch.long)
        sample[column] = torch.cat((input_tensor, pad_tensor))

    return sample


def pad(
    source: CheckpointableDataset,
    pad_values: Mapping[str, int],
    chunk_length: int,
) -> CheckpointableDataset:
    return source.map(
        functools.partial(
            _pad_fn,
            pad_values=pad_values,
            chunk_length=chunk_length,
        )
    )



from __future__ import annotations

from typing import Optional

from ...base import CheckpointableDataset, CheckpointableIterator, CollateFn, Sample, StateDict


class BatchIterator(CheckpointableIterator):
    def __init__(self, dataset: BatchDataset, source: CheckpointableIterator) -> None:
        self.dataset = dataset
        self.source = source

    def __next__(self) -> Sample:
        batch = []
        try:
            for _ in range(self.dataset.batch_size):
                batch.append(next(self.source))
        except StopIteration:
            if self.dataset.drop_last or not batch:
                raise

        return self.dataset.collate_fn(batch)

    def state_dict(self) -> StateDict:
        return self.source.state_dict()


class BatchDataset(CheckpointableDataset):
    def __init__(
        self,
        source: CheckpointableDataset,
        batch_size: int,
        collate_fn: CollateFn,
        drop_last: bool,
    ) -> None:
        self.source = source
        self.batch_size = batch_size
        self.collate_fn = collate_fn
        self.drop_last = drop_last

    def iter(self, state_dict: Optional[StateDict] = None) -> CheckpointableIterator:
        return BatchIterator(self, self.source.iter(state_dict=state_dict))



from __future__ import annotations

import random
from collections import deque
from logging import getLogger
from typing import Generic, Optional, Sequence, TypeVar

from ...base import CheckpointableIterator, FileFormat, Sample, StateDict
from .shard_reader import ShardReader


logger = getLogger()


T = TypeVar("T")


class EpochShuffleList(Generic[T]):
    def __init__(self, items: Sequence[T], shuffle: bool, seed: int) -> None:
        self.items = list(items)
        self.shuffle = shuffle
        self.rng = random.Random(seed)
        self.epoch = 0

        if shuffle:
            self.rng.shuffle(self.items)

    def __getitem__(self, index: int) -> T:
        return self.items[index]

    def __len__(self) -> int:
        return len(self.items)

    def advance_epoch(self, epoch: int) -> None:
        if self.epoch > epoch:
            raise ValueError(f"Cannot decrease epoch from {self.epoch} to {epoch}")

        while self.epoch < epoch:
            if self.shuffle:
                self.rng.shuffle(self.items)
            self.epoch += 1


class ShardsMux(CheckpointableIterator):
    def __init__(
        self,
        urls: Sequence[str],
        format: FileFormat,
        repeat: bool,
        shuffle: bool,
        n_active_shards: int,
        n_standby_shards: int,
        timeout: float,
        n_prefetch_samples: int,
        seed: int,
        next_active_shard: Optional[tuple[int, int]] = None,  # (epoch, index in epoch)
        active_shard_states: Optional[Sequence[StateDict]] = None,
    ) -> None:
        self.urls = EpochShuffleList(urls, shuffle=shuffle, seed=seed)
        self.format = format
        self.repeat = repeat
        self.shuffle = shuffle
        self.timeout = timeout
        self.n_prefetch_samples = n_prefetch_samples
        self.n_active_shards = min(n_active_shards, len(self.urls))

        # When we handle the boundary of epochs, we have smaller number of active shards.
        # So, instead of limiting the number of stand-by shards,
        # we limit the total number of open shards.
        self.n_open_shards = self.n_active_shards + n_standby_shards

        # We have `next_active_shard` as an argument, because it is more clear for checkpointing.
        # However, internally, it is more useful to manage `next_standby_shard`.
        # As the stand-by queue is initially empty, `next_standby_shard` is `next_active_shard`,
        # because the stand-by queue and we will refill the active shards from the stand-by queue.
        self.next_standby_shard = next_active_shard or (0, 0)

        self.active_shards = deque(
            ShardReader.from_state_dict(
                state_dict=state_dict, timeout=timeout, n_prefetch_samples=n_prefetch_samples
            )
            for state_dict in (active_shard_states or [])
        )
        self.standby_shards: deque[ShardReader] = deque()

        # Start the prefetching
        self._refill_active_shards()

    def _refill_standby_shards(self) -> None:
        while len(self.active_shards) + len(self.standby_shards) < self.n_open_shards:
            epoch, index = self.next_standby_shard
            if index >= len(self.urls):
                self.next_standby_shard = (epoch + 1, 0)
                continue
            if epoch > 0 and not self.repeat:
                break

            self.urls.advance_epoch(epoch)
            assert self.urls.epoch == epoch
            url = self.urls[index]

            logger.debug(f"New stand-by shard: {url}")
            shard_reader = ShardReader(
                url,
                self.format,
                timeout=self.timeout,
                n_prefetch_samples=self.n_prefetch_samples,
                n_samples_yielded=0,
                epoch=epoch,
                index_in_epoch=index,
            )
            self.standby_shards.append((shard_reader))
            self.next_standby_shard = (epoch, index + 1)

    def _refill_active_shards(self) -> None:
        self._refill_standby_shards()
        while len(self.active_shards) < self.n_active_shards:
            if len(self.standby_shards) == 0:
                break
            standby_shard = self.standby_shards[0]

            # We don't want to mix shards from different epochs
            if len(self.active_shards) == 0:
                epoch = None
            else:
                epoch = self.active_shards[0].epoch
            if epoch is not None and epoch != standby_shard.epoch:
                break

            # Move the shard from the stand-by queue to the active shards
            logger.debug(f"New active shard: {standby_shard.url}")
            self.standby_shards.popleft()
            self.active_shards.append(standby_shard)

            self._refill_standby_shards()

    def __next__(self) -> Sample:
        while True:
            self._refill_active_shards()

            # Done with all shards?
            if len(self.active_shards) == 0:
                raise StopIteration

            shard = self.active_shards.popleft()
            try:
                sample = next(shard)
                self.active_shards.append(shard)
                return sample
            except StopIteration:
                logger.debug(f"Complete shard: {shard.url}")
                continue

    @property
    def _next_active_shard(self) -> tuple[int, int]:
        if self.standby_shards:
            return (self.standby_shards[0].epoch, self.standby_shards[0].index_in_epoch)
        else:
            return self.next_standby_shard

    def state_dict(self) -> StateDict:
        return {
            "next_active_shard": self._next_active_shard,
            "active_shards": [active_shard.state_dict() for active_shard in self.active_shards],
        }



from __future__ import annotations

from typing import Optional

from ..base import CheckpointableDataset, CheckpointableIterator, Sample, StateDict


def generate_stratified_sampling_order(weights: list[float], length: int) -> list[int]:
    total_weights = sum(weights)
    weights = [weight / total_weights for weight in weights]

    order = []
    score = [0.0] * len(weights)
    for _ in range(length):
        for i, w in enumerate(weights):
            score[i] += w

        source_id = max(range(len(weights)), key=lambda i: score[i])
        score[source_id] -= 1.0
        order.append(source_id)

    return order


class InterleaveIterator(CheckpointableIterator):
    def __init__(
        self, dataset: InterleaveDataset, sources: list[CheckpointableIterator], start_index: int
    ) -> None:
        self.dataset = dataset
        self.sources = sources
        self.index = start_index
        self.source_ids = generate_stratified_sampling_order(
            self.dataset.weights, self.dataset.chunk_size
        )

    def __next__(self) -> Sample:
        source_id = self.source_ids[self.index % self.dataset.chunk_size]
        self.index += 1

        # TODO: we may want to have option to continue generation until all sources are exhausted
        # (currently we stop when the first source is exhausted)
        return next(self.sources[source_id])

    def state_dict(self) -> StateDict:
        state_dict: StateDict = {
            f"source{source_id}": source.state_dict()
            for source_id, source in enumerate(self.sources)
        }
        state_dict["index"] = self.index
        return state_dict


class InterleaveDataset(CheckpointableDataset):
    def __init__(
        self,
        sources: list[CheckpointableDataset],
        weights: Optional[list[float]],
        chunk_size: int = 1024,
    ):
        if weights is not None:
            if len(weights) != len(sources):
                raise ValueError(
                    f"Number of weights ({len(weights)}) must match number"
                    f" of sources ({len(sources)})"
                )
            if sum(weights) == 0:
                raise ValueError("Sum of weights must be greater than 0")

        self.sources = sources
        self.weights = weights or ([1] * len(sources))
        self.chunk_size = chunk_size

    def iter(self, state_dict: Optional[StateDict] = None) -> InterleaveIterator:
        if state_dict:
            source_state_dicts = [
                state_dict.pop(f"source{source_id}") for source_id in range(len(self.sources))
            ]
            index = state_dict.pop("index")
            if state_dict:
                raise ValueError(f"Unexpected keys in state_dict: {state_dict.keys()}")
        else:
            source_state_dicts = [None] * len(self.sources)
            index = 0

        iters = [
            source.iter(state_dict=state_dict)
            for source, state_dict in zip(self.sources, source_state_dicts)
        ]
        return InterleaveIterator(self, iters, start_index=index)


def interleave_datasets(
    sources: list[CheckpointableDataset],
    weights: Optional[list[float]] = None,
    chunk_size: int = 1024,
) -> CheckpointableDataset:
    return InterleaveDataset(sources, weights, chunk_size=chunk_size)



from __future__ import annotations

from typing import Optional, Sequence, Union

import numpy as np
import torch

from ...base import Sample, TokenArray


def tensor_from_token_array(data: Optional[Union[int, TokenArray]]) -> torch.Tensor:
    """
    Converts token arrays, typically the output from tokenizers, into a 1D PyTorch tensor.

    This function handles multiple input types including integers, torch tensors, and numpy arrays.
    The goal is to produce a consistent 1D tensor output for token data.

    Args:
        data (Optional[Union[int, TokenArray]]): Token data to be converted. It can be `None`,
        an integer, a torch tensor, or a numpy array.

    Returns:
        torch.Tensor: A 1-dimensional tensor representation of the input token data.

    Raises:
        ValueError: If the provided numpy array is not of an integer type.
        ValueError: If the resulting tensor is not of dtype `torch.long`.
        ValueError: If the tensor dimensions are not compatible with the expected shape.
    """
    if data is None:
        data = []
    if isinstance(data, int):
        data = [data]

    if isinstance(data, torch.Tensor):
        tensor = data
    elif isinstance(data, np.ndarray):
        if np.issubdtype(data.dtype, np.integer):
            tensor = torch.from_numpy(data).long()
        else:
            raise ValueError(f"Expected integer ndarray, got ndarray of {data.dtype}")
    else:
        tensor = torch.tensor(data, dtype=torch.long)

    if tensor.dtype != torch.long:
        raise ValueError(f"Expected long tensor, got {tensor.dtype}")

    if tensor.dim() == 2:
        if tensor.shape[0] != 1:
            raise ValueError(
                "input_ids must be 1-dimensional tensor or 2-dimensional tensor with batch size 1"
            )
        tensor = tensor[0]
    elif tensor.dim() != 1:
        raise ValueError("input_ids must be 1-dimensional tensor")
    return tensor


class TokensQueue:
    """
    A buffer dictionary class to manage, concatenate, pack, and chunk token arrays on the fly.
    """

    def __init__(self, columns: Sequence[str], buffers: Optional[dict[str, torch.Tensor]]) -> None:
        if len(columns) == 0:
            raise ValueError("columns must be non-empty")

        self.columns = columns

        if buffers is not None:
            if set(columns) != set(buffers.keys()):
                raise ValueError(
                    f"columns and buffers keys must match: {columns} != {buffers.keys()}"
                )
            self.buffers = buffers
        else:
            self.buffers = {column: torch.empty(0, dtype=torch.long) for column in columns}

    def length(self) -> int:
        return len(next(iter(self.buffers.values())))

    def tensor_dict_from_sample(self, sample: Sample) -> dict[str, torch.Tensor]:
        return {column: tensor_from_token_array(sample[column]) for column in self.columns}

    def push_from_tensor_dict(self, tensor_dict: dict[str, torch.Tensor]) -> None:
        input_length: Optional[int] = None
        for column in self.columns:
            tokens = tensor_dict[column]

            if input_length is None:
                input_length = len(tokens)
            else:
                if input_length != len(tokens):
                    raise ValueError("All columns must have the same length")

            self.buffers[column] = torch.cat([self.buffers[column], tokens])

    def push_from_sample(self, sample: Sample) -> None:
        self.push_from_tensor_dict(self.tensor_dict_from_sample(sample))

    def pop_by_length(self, length: int) -> Sample:
        output = {column: self.buffers[column][:length] for column in self.columns}
        self.buffers = {column: self.buffers[column][length:] for column in self.columns}
        return output

    def pop_all(self) -> Sample:
        output = {column: self.buffers[column] for column in self.columns}
        self.buffers = {column: torch.empty(0, dtype=torch.long) for column in self.columns}
        return output



from __future__ import annotations

import json
import threading
from logging import getLogger
from queue import Empty, Queue
from typing import IO, Any, Generator, Optional

import smart_open

from ...base import FileFormat, Sample


logger = getLogger(__name__)


class ReaderThreadResult:
    error: Optional[Exception]
    result: Optional[Sample]
    end: bool

    def __init__(
        self, error: Optional[Exception] = None, result: Optional[Sample] = None, end: bool = False
    ) -> None:
        self.error = error
        self.result = result
        self.end = end


def _deduce_format(url: str) -> FileFormat:
    url = url.lower()
    if url.endswith(".cbor"):
        return "cbor"
    elif url.endswith(".jsonl"):
        return "jsonl"
    else:
        raise ValueError(f"Could not deduce format from url: {url}")


def _is_text(format: FileFormat) -> bool:
    return format in ["jsonl"]


def _generate_from_stream_jsonl(
    url: str,
    stream: IO[str],
    n_samples_to_skip: int,
    max_consecutive_errors: int = 10,
) -> Generator[Sample, None, None]:
    line_count = 0
    n_consecutive_errors = 0
    while True:
        line = stream.readline()

        # EOF
        if not line:
            break
        line_count += 1

        # Empty line
        line = line.strip()
        if not line:
            continue

        # Decode and yield
        try:
            sample = json.loads(line)
            n_consecutive_errors = 0

            if n_samples_to_skip > 0:
                n_samples_to_skip -= 1
            else:
                yield sample
        except json.decoder.JSONDecodeError:
            n_consecutive_errors += 1
            logger.exception(
                f"Decoding error encountered on line {line_count} of URL {url}. "
                f"Skipping this line and moving on to the next. "
                f"This is error {n_consecutive_errors}/{max_consecutive_errors} in a row. "
                f"If {max_consecutive_errors} consecutive errors occur, it will be aborted."
            )
            logger.debug(
                f"Content of the problematic line (line={line_count}, url={url}):\n"
                f"{'-' * 80}\n"
                f"{line}\n"
                f"{'-' * 80}\n"
            )
            if n_consecutive_errors >= max_consecutive_errors:
                # TODO: we probably don't need to retry for this error
                raise ValueError(
                    f"Encountered {max_consecutive_errors} consecutive decoding errors "
                    f"for URL {url}. Terminating the generator."
                )

    logger.debug(f"JSONL EOF: line={line_count}, url={url}")


def _generate_form_stream_cbor(
    url: str,
    stream: IO[bytes],
    n_samples_to_skip: int,
) -> Generator[Sample, None, None]:
    import cbor2

    try:
        while True:
            sample = cbor2.load(stream)

            if n_samples_to_skip > 0:
                n_samples_to_skip -= 1
            else:
                yield sample
    except EOFError:
        pass


def _generate_from_stream(
    url: str,
    format: FileFormat,
    stream: IO[Any],
    n_samples_to_skip: int,
) -> Generator[Sample, None, None]:
    if format == "jsonl":
        yield from _generate_from_stream_jsonl(url, stream, n_samples_to_skip)
    elif format == "cbor":
        yield from _generate_form_stream_cbor(url, stream, n_samples_to_skip)
    else:
        raise ValueError(f"Unknwon format: {format}")


def _reader_thread(
    url: str,
    format: FileFormat,
    n_samples_to_skip: int,
    queue: Queue[ReaderThreadResult],
    finish_event: threading.Event,
) -> None:
    mode = "r" if _is_text(format) else "rb"

    try:
        logger.debug(f"Read thread starting: url={url}")
        with smart_open.open(url, mode) as f:
            gen = _generate_from_stream(
                url=url,
                format=format,
                stream=f,
                n_samples_to_skip=n_samples_to_skip,
            )
            for sample in gen:
                if finish_event.is_set():
                    logger.debug(f"Read thread interrupted: url={url}")
                    return
                queue.put(ReaderThreadResult(result=sample))

        logger.debug(f"Read thread completed: url={url}")
        queue.put(ReaderThreadResult(end=True))
    except Exception as e:
        logger.exception(f"Exception in read thread: url={url}")
        queue.put(ReaderThreadResult(error=e))


def _generator(
    thread: threading.Thread,
    queue: Queue[ReaderThreadResult],
    finish_event: threading.Event,
    timeout: float,
) -> Generator[Sample, None, None]:
    try:
        while True:
            result = queue.get(timeout=timeout)
            if result.end:
                break
            elif result.error:
                raise result.error
            else:
                assert result.result is not None
                yield result.result
    finally:
        finish_event.set()
        try:
            while True:
                queue.get_nowait()
        except Empty:
            pass
        thread.join()


def yield_samples(
    url: str,
    format: FileFormat,
    n_samples_to_skip: int,
    n_prefetch_samples: int,
    timeout: float,
) -> Generator[Sample, None, None]:
    if format == "auto":
        format = _deduce_format(url)
    format = format.lower()  # type: ignore

    queue: Queue[ReaderThreadResult] = Queue(n_prefetch_samples)
    finish_event = threading.Event()

    thread = threading.Thread(
        target=_reader_thread,
        daemon=True,
        kwargs={
            "url": url,
            "format": format,
            "n_samples_to_skip": n_samples_to_skip,
            "queue": queue,
            "finish_event": finish_event,
        },
    )
    thread.start()

    # `_generator` should be a different method because we want to launch the thread
    # before `next` is called on the generator for the first time.
    return _generator(thread, queue, finish_event, timeout)



from __future__ import annotations

from typing import Optional

from ...base import CheckpointableDataset, CheckpointableIterator, Sample, StateDict


class CountIterator(CheckpointableIterator):
    def __init__(
        self, dataset: CountDataset, source: CheckpointableIterator, start_count: int
    ) -> None:
        self.count = start_count
        self.dataset = dataset
        self.source = source

    def __next__(self) -> Sample:
        count = self.count
        self.count += 1
        if self.dataset.max_count is not None and self.count > self.dataset.max_count:
            raise StopIteration

        sample = next(self.source)

        if self.dataset.count_column:
            sample[self.dataset.count_column] = count

        return sample

    def state_dict(self) -> StateDict:
        return {
            "source": self.source.state_dict(),
            "count": self.count,
        }


class CountDataset(CheckpointableDataset):
    def __init__(
        self,
        source: CheckpointableDataset,
        count_column: Optional[str] = None,
        max_count: Optional[int] = None,
    ) -> None:
        self.source = source
        self.count_column = count_column
        self.max_count = max_count

    def iter(self, state_dict: Optional[StateDict] = None) -> CheckpointableIterator:
        if state_dict:
            source_state_dict = state_dict.pop("source")
            count = state_dict.pop("count")
            if state_dict:
                raise ValueError(f"Unexpected keys in state_dict: {state_dict.keys()}")
        else:
            source_state_dict = None
            count = 0

        return CountIterator(self, self.source.iter(state_dict=source_state_dict), count)



from __future__ import annotations

from typing import Optional, Sequence, Union

import braceexpand

from ...base import CheckpointableDataset, CheckpointableIterator, FileFormat, StateDict
from .shards_mux import ShardsMux


class FilesDataset(CheckpointableDataset):
    def __init__(
        self,
        urls: Union[str, Sequence[str]],
        format: FileFormat,
        repeat: bool,
        shuffle_shards: bool,
        n_active_shards: int,
        n_standby_shards: int,
        timeout: float,
        n_prefetch_samples: int,
        seed: int,
    ) -> None:
        if isinstance(urls, str):
            urls = [urls]
        self.urls: list[str] = sum((list(braceexpand.braceexpand(url)) for url in urls), [])

        self.format = format
        self.n_acive_shards = n_active_shards
        self.n_standby_shards = n_standby_shards
        self.timeout = timeout
        self.n_prefetch_samples = n_prefetch_samples
        self.repeat = repeat
        self.shuffle_shards = shuffle_shards
        self.seed = seed

    def iter(self, state_dict: Optional[StateDict] = None) -> CheckpointableIterator:
        if state_dict is not None:
            next_active_shard = state_dict.pop("next_active_shard")
            active_shard_states = state_dict.pop("active_shards")
            if state_dict:
                raise ValueError(f"Unexpected keys in state_dict: {state_dict.keys()}")
        else:
            next_active_shard = (0, 0)
            active_shard_states = None

        return ShardsMux(
            urls=self.urls,
            format=self.format,
            repeat=self.repeat,
            shuffle=self.shuffle_shards,
            n_active_shards=self.n_acive_shards,
            n_standby_shards=self.n_standby_shards,
            timeout=self.timeout,
            n_prefetch_samples=self.n_prefetch_samples,
            seed=self.seed,
            next_active_shard=next_active_shard,
            active_shard_states=active_shard_states,
        )



from __future__ import annotations

from typing import Optional, Sequence

from ..base import CheckpointableDataset, CheckpointableIterator, Sample, StateDict


class ConcatIterator(CheckpointableIterator):
    def __init__(
        self, dataset: ConcatDataset, source_iter: CheckpointableIterator, source_index: int
    ) -> None:
        self.dataset = dataset
        self.source_iter = source_iter
        self.source_index = source_index

    def __next__(self) -> Sample:
        while self.source_index < len(self.dataset.sources):
            try:
                return next(self.source_iter)
            except StopIteration:
                self.source_index += 1
                if self.source_index >= len(self.dataset.sources):
                    break
                self.source_iter = self.dataset.sources[self.source_index].iter()
        raise StopIteration()

    def state_dict(self) -> StateDict:
        return {
            "source_iter": self.source_iter.state_dict(),
            "source_index": self.source_index,
        }


class ConcatDataset(CheckpointableDataset):
    def __init__(self, sources: Sequence[CheckpointableDataset]) -> None:
        if len(sources) == 0:
            raise ValueError("ConcatDataset must have at least one source")
        self.sources = sources

    def iter(self, state_dict: Optional[StateDict] = None) -> CheckpointableIterator:
        if state_dict:
            source_index = state_dict.pop("source_index")
            if source_index >= len(self.sources):
                raise ValueError(
                    f"source_index out of range: {source_index} >= {len(self.sources)}"
                )
            source_iter_state = state_dict.pop("source_iter")
            if state_dict:
                raise ValueError(f"Unexpected keys in state_dict: {state_dict.keys()}")
        else:
            source_index = 0
            source_iter_state = None

        source = self.sources[source_index]
        source_iter = source.iter(source_iter_state)
        return ConcatIterator(self, source_iter, source_index)


def concat_datasets(sources: Sequence[CheckpointableDataset]) -> ConcatDataset:
    return ConcatDataset(sources)



from .batch import BatchDataset
from .count import CountDataset
from .filter_map import FilterMapDataset
from .parallel_filter_map import ParallelFilterMapDataset
from .shuffle import ShuffleDataset
from .stride import StrideDataset


__all__ = [
    "BatchDataset",
    "CountDataset",
    "FilterMapDataset",
    "ParallelFilterMapDataset",
    "ShuffleDataset",
    "StrideDataset",
]



from __future__ import annotations

import warnings
import weakref
from typing import TYPE_CHECKING, Any, Optional

from ..base import CheckpointableDataset, CheckpointableIterator, Sample, StateDict


if TYPE_CHECKING:
    import streaming


class MosaicmlIterator(CheckpointableIterator):
    def __init__(self, dataset: MosaicmlDataset, start_index: int) -> None:
        self.dataset = dataset
        self.iter = iter(dataset.source)
        self.index = start_index

    def __next__(self) -> Sample:
        self.index += 1

        try:
            sample = next(self.iter)
        except StopIteration:
            if not self.dataset.repeat:
                raise

            # Starting the next epoch
            self.iter = iter(self.dataset.source)
            sample = next(self.iter)

        return sample

    def state_dict(self) -> StateDict:
        # The StreamingDataset in MosaicML is semi-stateful. It remembers the epoch, but it doesn't
        # remember the position within an epoch. We need to provide the position within the epoch.
        index_in_epoch = self.index % len(self.dataset.source)
        return {
            "index": self.index,
            "dataset": self.dataset.source.state_dict(index_in_epoch, from_beginning=True),
        }


class MosaicmlDataset(CheckpointableDataset):
    def __init__(self, source: streaming.StreamingDataset, repeat: bool) -> None:
        self.source = source
        self.repeat = repeat

        self.iter_called = False
        self.iters: weakref.WeakSet[MosaicmlIterator] = weakref.WeakSet()

    def iter(self, state_dict: Optional[dict[str, Any]] = None) -> CheckpointableIterator:
        if self.iters:
            msg = (
                "The existing MosaicmlIterator for this MosaicmlDataset is still alive. If you're "
                "trying to handle multiple MosaicmlIterators for a single MosaicmlDataset, there "
                "could be unintended behavior. This is because, unfortunately, "
                "the StreamingDataset in MosaicML is somewhat stateful, and the Iterators may "
                "affect each other's behavior."
            )
            warnings.warn(msg)

        if state_dict:
            if self.iter_called:
                msg = (
                    "At least in the current version (0.5.1), in order to resume a MosaicML's "
                    "StreamingDataset with a state_dict, it looks that it must not have accessed "
                    "the data before. The state_dict is ignored for the subsequent calls."
                )
                raise ValueError(msg)

            self.source.load_state_dict(state_dict["dataset"])
            index = state_dict["index"]
        else:
            self.source.next_epoch = 0
            index = 0

        self.iter_called = True
        iterator = MosaicmlIterator(self, start_index=index)
        self.iters.add(iterator)
        return iterator



from __future__ import annotations

import random
from typing import Optional

from ...base import CheckpointableDataset, CheckpointableIterator, Sample, StateDict


class ShuffleIterator(CheckpointableIterator):
    def __init__(
        self,
        source: CheckpointableIterator,
        buffer_size: int,
        rng: random.Random,
        buffer: list[Sample],
    ) -> None:
        self.source = source
        self.buffer_size = buffer_size
        self.rng = rng
        self.buffer = buffer

    def _refill_buffer(self) -> None:
        while len(self.buffer) < self.buffer_size:
            try:
                self.buffer.append(next(self.source))
            except StopIteration:
                break

    def __next__(self) -> Sample:
        self._refill_buffer()
        if len(self.buffer) == 0:
            raise StopIteration()

        index = self.rng.randrange(len(self.buffer))
        self.buffer[index], self.buffer[-1] = self.buffer[-1], self.buffer[index]
        return self.buffer.pop()

    def state_dict(self) -> StateDict:
        return {
            "source": self.source.state_dict(),
            "buffer": self.buffer.copy(),
            "rng": self.rng.getstate(),
        }


class ShuffleDataset(CheckpointableDataset):
    def __init__(self, source: CheckpointableDataset, buffer_size: int, seed: int) -> None:
        if buffer_size < 1:
            raise ValueError(f"Invalid buffer_size (got {buffer_size}, expected >= 1)")

        self.source = source
        self.buffer_size = buffer_size
        self.seed = seed

    def iter(self, state_dict: Optional[StateDict] = None) -> CheckpointableIterator:
        rng = random.Random(self.seed)
        if state_dict is not None:
            source = state_dict.pop("source")
            buffer = state_dict.pop("buffer")
            rng.setstate(state_dict.pop("rng"))
            if state_dict:
                raise ValueError("Unexpected state_dict keys: {}".format(state_dict.keys()))
        else:
            source = None
            buffer = []

        return ShuffleIterator(
            source=self.source.iter(source),
            buffer_size=self.buffer_size,
            rng=rng,
            buffer=buffer,
        )



from typing import Iterable, Iterator, Optional

from ..base import CheckpointableDataset, CheckpointableIterator, Sample, StateDict


class IterableIterator(CheckpointableIterator):
    def __init__(self, dataset: "IterableDataset", start_index: int) -> None:
        self.dataset = dataset
        self.index = start_index
        self.source_iter = self._initialize_source_iter(start_index)

    def _initialize_source_iter(self, start_index: int) -> Iterator[Sample]:
        it = iter(self.dataset.iterable)

        for _ in range(start_index):
            try:
                next(it)
            except StopIteration:
                raise ValueError(f"start_index larger than iterable length: {start_index}")

        return it

    def __next__(self) -> Sample:
        try:
            sample = next(self.source_iter)
            self.index += 1
            return sample
        except StopIteration:
            if not self.dataset.repeat:
                raise StopIteration()

            self.source_iter = iter(self.dataset.iterable)
            sample = next(self.source_iter)
            self.index = 1
            return sample

    def state_dict(self) -> StateDict:
        return {
            "index": self.index,
        }


class IterableDataset(CheckpointableDataset):
    def __init__(
        self,
        iterable: Iterable[Sample],
        repeat: bool,
    ) -> None:
        self.iterable = iterable
        self.repeat = repeat

    def iter(self, state_dict: Optional[StateDict] = None) -> CheckpointableIterator:
        if state_dict:
            if set(state_dict.keys()) != {"index"}:
                raise ValueError(
                    f"Invalid state_dict (expect keys: 'index', keys: {state_dict.keys()})"
                )
            index = state_dict["index"]
        else:
            index = 0

        return IterableIterator(self, start_index=index)



import random
from typing import Optional, Sequence

from ..base import CheckpointableDataset, CheckpointableIterator, Sample, StateDict


class SequenceIterator(CheckpointableIterator):
    def __init__(self, dataset: "SequenceDataset", start_index: int) -> None:
        self.dataset = dataset
        self.index = start_index
        self.epoch = 0
        self.rng = random.Random(dataset.shuffle_seed)
        self.order = list(range(len(self.dataset.sequence)))
        self._reshuffle()

    def _reshuffle(self) -> None:
        if self.dataset.shuffle_per_epoch:
            self.rng.shuffle(self.order)

    def __next__(self) -> Sample:
        while True:
            index_in_epoch = self.index - self.epoch * len(self.dataset.sequence)
            if index_in_epoch < len(self.dataset.sequence):
                break

            if not self.dataset.repeat:
                raise StopIteration()

            self.epoch += 1
            self._reshuffle()

        sample = self.dataset.sequence[self.order[index_in_epoch]]
        self.index += 1
        return sample

    def state_dict(self) -> StateDict:
        return {
            "index": self.index,
        }


class SequenceDataset(CheckpointableDataset):
    def __init__(
        self,
        sequence: Sequence[Sample],
        repeat: bool,
        shuffle: bool,
        shuffle_seed: int,
    ) -> None:
        self.sequence = sequence
        self.repeat = repeat
        self.shuffle_per_epoch = shuffle
        self.shuffle_seed = shuffle_seed

    def iter(self, state_dict: Optional[StateDict] = None) -> CheckpointableIterator:
        if state_dict:
            if set(state_dict.keys()) != {"index"}:
                raise ValueError(
                    f"Invalid state_dict (expect keys: 'index', keys: {state_dict.keys()})"
                )
            index = state_dict["index"]
        else:
            index = 0

        return SequenceIterator(self, start_index=index)



from .basic import (
    BatchDataset,
    CountDataset,
    FilterMapDataset,
    ParallelFilterMapDataset,
    ShuffleDataset,
    StrideDataset,
)
from .language_modeling import (
    ChunkDataset,
    ConcatChunkDataset,
    PackChunkDataset,
    add_bos_eos,
    ensure_bos_eos,
    pad,
    tokenize,
)


__all__ = [
    "FilterMapDataset",
    "FilterMapFn",
    "CountDataset",
    "BatchDataset",
    "ParallelFilterMapDataset",
    "ShuffleDataset",
    "StrideDataset",
    "ChunkDataset",
    "ConcatChunkDataset",
    "PackChunkDataset",
    "add_bos_eos",
    "ensure_bos_eos",
    "pad",
    "tokenize",
]
